#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define True  (1)
#define False (0)

/* for check/debug */
#define CHECK True

/* prototypes */
void lorentz(double v[4],double psi,double theta,double phi,double al[4][4]);
void prodmm(double a[4][4], double b[4][4], double r[4][4]);
void prodmv(double a[4][4], double v[4], double r[4]);
void prmat(char *str, double ltr[4][4]);
void prvect(char *str, double lv[4]);

/*----------------------------------------------------------------*/
int main(int argc, char **argv)
{
    double pi, phi, psi, theta;
    double v[4];
    double al[4][4];
    double p[4], pb[4];
    double q[4], qb[4];

    /* constant */
    pi = 2*asin(1.0e0);

    /* input data */
    phi   =   0.30e0 * pi;
    psi   =   0.40e0 * pi;
    theta =   0.50e0 * pi;
    v[1]  = 0.2e0;
    v[2]  = 0.3e0;
    v[3]  = 0.5e0;

    /* four momentum for test */
    p[0]  = 1;
    p[1]  = 0;
    p[2]  = 0;
    p[3]  = 0;

    q[0]  = 4;
    q[1]  = 1;
    q[2]  = 2;
    q[3]  = 3;

    /* construct the matrix representation "al" of Lorentz transformation */
    lorentz(v, psi, theta, phi, al);

    /* Lorentz transformation of four vectors */

    prodmv(al, p,  pb);

    prvect("p", p);
    prvect("al*p", pb);

    prodmv(al, q,  qb);

    prvect("q", q);
    prvect("al*q", qb);

    return 0;
}

/*----------------------------------------------------------------
 *   Construct the matrix of Lorentz transformation : al = boost * rotation
 *
 *     Input :
 *       double v[3) : Space components of transformed four vector
 *                     of (1,0,0,0).
 *       double psi, theta, phi : Eular angle.
 *
 *     Output :
 *       double al[4][4] : resulting matrix.
 */
void lorentz(double v[4], double psi, double theta, double phi,
             double al[4][4])
{
    /* local variables */
    double rot1[4][4], rot2[4][4], rot3[4][4];
    double rot[4][4],  rot21[4][4];
    double boost[4][4], g[4], br1[4][4];
    double vsq, vd;

    int k, k1, k2, j;

    /* clear matrix */
    for(k = 0; k < 4; k++) {
        for(j = 0; j < 4; j++) {
            rot1[j][k]  = 0;
            rot2[j][k]  = 0;
            rot3[j][k]  = 0;
            boost[j][k] = 0;
        }
    }
    for(k = 0; k < 4; k++) {
        rot1[k][k]  = 1;
        rot2[k][k]  = 1;
        rot3[k][k]  = 1;
        boost[k][k] = 1;
    }

    rot1[1][1] =   cos(phi);
    rot1[1][2] = - sin(phi);
    rot1[2][1] =   sin(phi);
    rot1[2][2] =   cos(phi);

    rot2[1][1] =   cos(theta);
    rot2[1][3] = - sin(theta);
    rot2[3][1] =   sin(theta);
    rot2[3][3] =   cos(theta);

    rot3[1][1] =   cos(psi);
    rot3[1][2] = - sin(psi);
    rot3[2][1] =   sin(psi);
    rot3[2][2] =   cos(psi);

    vsq = v[1]*v[1] + v[2]*v[2] + v[3]*v[3];
    vd  = sqrt(1+vsq);
    boost[0][0] = vd;
    for(k = 1; k < 4; k++) {
        boost[0][k] = v[k];
        boost[k][0] = v[k];
    }
    for(k = 1; k < 4; k++) {
        for(j = 1; j < 4; j++) {
            boost[j][k] = (vd-1)*v[j]*v[k]/vsq;
        }
        boost[k][k] += 1;
    }

    prodmm(rot2,  rot1, rot21);
    prodmm(rot21, rot3, rot);
    prodmm(boost, rot,  al);

    /* for check */
#if CHECK
        g[0] =  1;
        g[1] = -1;
        g[2] = -1;
        g[3] = -1;

        for(k1 = 0; k1 < 4; k1++) {
            for(k2 = 0; k2 < 4; k2++) {
                br1[k1][k2] = 0;
                for(j = 0; j < 4; j++) {
                    br1[k1][k2] += g[j]*al[j][k1]*al[j][k2];
                }
            }
          }
  
        prmat("rot1",  rot1);
        prmat("rot2",  rot2);
        prmat("rot3",  rot3);
        prmat("boost", boost);
        prmat("rot2*rot1", rot21);
        prmat("rot3*rot2*rot1", rot);
        prmat("al = boost*rot3*rot2*rot1", al);
        prmat("g : check", br1);
#endif
}

/*----------------------------------------------------------------
 *   matrix * matrix : r = a*b
 *
 *   Input:
 *     double a[4][4], b[4][4] : input matrices
 *
 *   Ouput:
 *     double r[4][4] : output matrix
 *                          memory space of "r" should not be overlapped with
 *                          "a" nor "b".
 */
void prodmm(double a[4][4], double b[4][4], double r[4][4])
{
    int k1, k2, j;

    for(k1 = 0; k1 < 4; k1++) {
        for(k2 = 0; k2 < 4; k2++) {
            r[k1][k2] = 0;
            for(j = 0; j < 4; j++) {
                r[k1][k2] += a[k1][j]*b[j][k2];
            }
        }
    }
}

/*----------------------------------------------------------------
 *   matrix * vector : r = a*v
 *
 *   Input:
 *     double a[4][4] : input matrix
 *     double v[4]      : input vector
 *
 *   Output:
 *     double r[4]      : output vector
 *                          memory space of "r" should not be overlapped with
 *                          "a" nor "v".
 */
void prodmv(double a[4][4], double v[4], double r[4])
{
    int k, j;

    for(k = 0; k < 4; k++) {
        r[k] = 0;
        for(j = 0; j < 4; j++) {
            r[k] += a[k][j]*v[j];
        }
    }
}

/*----------------------------------------------------------------
 *   Print matrix "ltr" with message "str"
 *
 *   Input
 *     char   *str    : message character string
 *     double ltr[4][4] : matrix to be printed
 */
void prmat(char *str, double ltr[4][4])
{
    int j, k;

    printf("+++ %s +++\n ", str);
    for(j = 0; j < 4; j++) {
        for(k = 0; k < 4; k++) {
            printf(" %9.6f", ltr[j][k]);
        }
        printf("\n ");
    }
}

/*----------------------------------------------------------------
 *   Print four vector "lv" with message "str"
 *
 *   Input
 *     char   *str    : message character string
 *     double lv[4]       : four vector to be printed
 */
void prvect(char *str, double lv[4])
{
    double msq;
    int j;

    msq = lv[0]*lv[0] - lv[1]*lv[1] - lv[2]*lv[2] - lv[3]*lv[3];
    printf("+++ %s +++\n ", str);
    for(j = 0; j < 4; j++) {
        printf(" %9.6f", lv[j]);
    }
    printf(" m^2 = %9.6f\n ", msq);
}
