*   Example of Lorentz transformation
*---+----1----+----2----+----3----+----4----+----5----+----6----+----7--
      implicit none

      real*8 pi, phi, psi, theta
      real*8 v(3)
      real*8 al(0:3, 0:3)
      real*8 p(0:3), pb(0:3)
      real*8 q(0:3), qb(0:3)
*---+----1----+----2----+----3----+----4----+----5----+----6----+----7--

*  -- constant
      pi = 2*asin(1.0d0)

*  -- input data
      phi   =   0.30d0 * pi
      psi   =   0.40d0 * pi
      theta =   0.50d0 * pi
      v(1)  = 0.2d0
      v(2)  = 0.3d0
      v(3)  = 0.5d0

*  -- four momentum for test
      p(0)  = 1
      p(1)  = 0
      p(2)  = 0
      p(3)  = 0

      q(0)  = 4
      q(1)  = 1
      q(2)  = 2
      q(3)  = 3

*  -- construct the matrix representation 'al' of Lorentz transformation

      call lorentz(v, psi, theta, phi, al)

*  -- Lorentz transformation of four vectors

      call prodmv(al, p,  pb)

      call prvect('p', p)
      call prvect('al*p', pb)

      call prodmv(al, q,  qb)

      call prvect('q', q)
      call prvect('al*q', qb)

      end
*---+----1----+----2----+----3----+----4----+----5----+----6----+----7--
*   Construct the matrix of Lorentz transformation : al = boost * rotation
*
*     Input :
*       real*8 v(3) : Space components of transformed four vector
*                     of (1,0,0,0).
*       real*8 psi, theta, phi : Eular angle.
*
*     Output :
*       real*8 al(0:3, 0:3) : resulting matrix.
*
      subroutine lorentz(v, psi, theta, phi, al)
      implicit none

*  -- arguments
      real*8 v(3), phi, psi, theta, al(0:3, 0:3)

*  -- for check/debug
      logical check
      parameter (check = .true.)
c     parameter (check = .false.)

*  -- local variables
      real*8 rot1(0:3, 0:3),  rot2(0:3, 0:3), rot3(0:3, 0:3)
      real*8 rot(0:3, 0:3),   rot21(0:3, 0:3)
      real*8 boost(0:3, 0:3), g(0:3), br1(0:3, 0:3)
      real*8 vsq, vd

      integer k, k1, k2, j

*  -- clear matrix
      do k = 0, 3
        do j = 0, 3
          rot1(j,k)  = 0
          rot2(j,k)  = 0
          rot3(j,k)  = 0
          boost(j,k) = 0
        enddo
      enddo
      do k = 0, 3
        rot1(k,k)  = 1
        rot2(k,k)  = 1
        rot3(k,k)  = 1
        boost(k,k) = 1
      enddo

      rot1(1,1) =   cos(phi)
      rot1(1,2) = - sin(phi)
      rot1(2,1) =   sin(phi)
      rot1(2,2) =   cos(phi)

      rot2(1,1) =   cos(theta)
      rot2(1,3) = - sin(theta)
      rot2(3,1) =   sin(theta)
      rot2(3,3) =   cos(theta)

      rot3(1,1) =   cos(psi)
      rot3(1,2) = - sin(psi)
      rot3(2,1) =   sin(psi)
      rot3(2,2) =   cos(psi)

      vsq = v(1)**2 + v(2)**2 + v(3)**2
      vd = sqrt(1+vsq)
      boost(0, 0) = vd
      do k = 1, 3
        boost(0, k) = v(k)
        boost(k, 0) = v(k)
      enddo
      do k = 1, 3
        do j = 1, 3
          boost(j, k) = (vd-1)*v(j)*v(k)/vsq
        enddo
        boost(k,k) = boost(k,k) + 1
      enddo

      call prodmm(rot2,  rot1, rot21)
      call prodmm(rot21, rot3, rot)
      call prodmm(boost, rot,  al)

*  -- for check
      if(check) then
        g(0) =  1
        g(1) = -1
        g(2) = -1
        g(3) = -1

        do k1 = 0, 3
          do k2 = 0, 3
            br1(k1, k2) = 0
            do j = 0, 3
              br1(k1, k2) = br1(k1, k2) + g(j)*al(j,k1)*al(j,k2)
            enddo
          enddo
        enddo
  
        call prmat('rot1',  rot1)
        call prmat('rot2',  rot2)
        call prmat('rot3',  rot3)
        call prmat('boost', boost)
        call prmat('rot2*rot1', rot21)
        call prmat('rot3*rot2*rot1', rot)
        call prmat('al = boost*rot3*rot2*rot1', al)
        call prmat('g : check', br1)
      endif

      return
      end
*---+----1----+----2----+----3----+----4----+----5----+----6----+----7--
*   matrix * matrix : r = a*b
*
*   Input:
*     real*8 a(0:3, 0:3), b(0:3, 0:3) : input matrices
*
*   Ouput:
*     real*8 r(0:3, 0:3) : output matrix
*                          memory space of 'r' should not be overlapped with
*                          'a' nor 'b'.
*
      subroutine prodmm(a, b, r)
      implicit none
      real*8 a(0:3, 0:3), b(0:3, 0:3), r(0:3, 0:3)
      integer k1, k2, j

      do k1 = 0, 3
        do k2 = 0, 3
          r(k1,k2) = 0
          do j = 0, 3
            r(k1,k2) = r(k1,k2) + a(k1,j)*b(j,k2)
          enddo
        enddo
      enddo
      return
      end
*---+----1----+----2----+----3----+----4----+----5----+----6----+----7--
*   matrix * vector : r = a*v
*
*   Input:
*     real*8 a(0:3, 0:3) : input matrix
*     real*8 v(0:3)      : input vector
*
*   Output:
*     real*8 r(0:3)      : output vector
*                          memory space of 'r' should not be overlapped with
*                          'a' nor 'v'.
*
      subroutine prodmv(a, v, r)
      implicit none
      real*8 a(0:3, 0:3), v(0:3), r(0:3)
      integer k, j

      do k = 0, 3
        r(k) = 0
        do j = 0, 3
          r(k) = r(k) + a(k,j)*v(j)
        enddo
      enddo
      return
      end
*---+----1----+----2----+----3----+----4----+----5----+----6----+----7--
*   Print matrix 'ltr' with message 'str'
*
*   Input
*     character*(*) str    : message character string
*     real*8 ltr(0:3, 0:3) : matrix to be printed
*
      subroutine prmat(str, ltr)
      implicit none
      character*(*) str
      real*8 ltr(0:3, 0:3)

      integer j, k

      write(*,*) '+++ ', str, ' +++'
      do j = 0, 3
        write(*,'(1x,4f10.6)') (ltr(j, k), k=0,3)
      enddo

      return
      end
*---+----1----+----2----+----3----+----4----+----5----+----6----+----7--
*   Print four vector 'lv' with message 'str'
*
*   Input
*     character*(*) str    : message character string
*     real*8 lv(0:3)       : four vector to be printed
*
      subroutine prvect(str, lv)
      implicit none
      character*(*) str
      real*8 lv(0:3)

      real*8 msq
      integer j

      msq = lv(0)**2 - lv(1)**2 - lv(2)**2 - lv(3)**2
      write(*,*) '+++ ', str, ' +++'
      write(*,'(1x,4f10.6, a, f10.6)') (lv(j), j=0,3),' m^2 =',msq

      return
      end
