*   Integration by trapezoidal method
*   
*       /1
*       |  fnc(x) dx
*       /0
*
*      fncl(x) = x**npx * log(x)**npl
*      fnce(x) = x**npx * exp(npl * x)
*
*    Parameters 'npx' and 'npl' are passed through common block 'idat'.
*
*---+----1----+----2----+----3----+----4----+----5----+----6----+----7--
      implicit none

*  -- constants
      double precision three, four
      parameter (three=3, four=4)

*  -- the maximum iterations
      integer    maxj
      parameter (maxj = 25)

*  -- condition to finish iterations
*  -- require at least 'itmin' iterations
      integer    itmin
      parameter (itmin = 5)
*  -- require estimated error be less than 'ermin'
      double precision ermin
      parameter (ermin = 1.0d-8)

*  -- parameters passed to fncl and fnce for selecting integrand
      integer         npx, npl
      common    /idat/npx, npl

*  -- 'fnc' is a name of function
      external fncl, fnce
      double precision fncl, fnce
      double precision fncla, fncea

*  -- local variables
      integer          j, nc, n, maxn, ncc
      double precision r0, r1, rp0, rp1, tv
      double precision e0, e1, ep0, ep1, er0, er1, ans
*-----------------------------------------------------------------------
*  -- power of x
      npx  = 2

*  -- sweep parameters
      maxn = 10
      do n = -maxn, maxn

*    -- power of log/exp
        npl = n

*    -- analytic result
        if(npl.lt.0) then
          tv = fncla()
        else
          tv = fncea()
        endif

*    -- print messages
        write(*,*) 'integration from 0 to 1 of'
        if(npl.ge.0) then
          write(*,*) '    x**',npx,' * log(x)**',npl
        else
          write(*,*) '    x**',npx,' * exp(',npl,'* x)'
        endif
        write(*,*) maxj,'iterations.'

        write(*,*) '2**n,  integ,  err, err'

*    -- initial values for iterations
        r0 = 0
        nc = -1
        e0 = 100
        e1 = 100
        ncc = -1

*     do iterations
        do j = 0, maxj

*      -- save the previous results
          rp0 = r0
          rp1 = r1

*      -- estimation for 2**('nc'+1) points using previous result in 'r0'
          if(n .gt. 0) then
            call integt(nc, r0, fncl)
          else
            call integt(nc, r0, fnce)
          endif

*      -- for check
          if(j.ne.nc) then
            write(*,*) 'j = ', j, '!= nc=',nc
          endif

*      -- for acceleration
          r1 = (four*r0 - rp0)/three

*      -- previous errors
          ep0 = e0
          ep1 = e1

*      -- new errors
          e0 = abs(rp0-r0)
          e1 = abs(rp1-r1)

*      -- ration of new/previous
          if(ep0.eq.0) then
            er0 = e0
          else
            er0 = e0/ep0
          endif
          if(ep1.eq.0) then
            er1 = e1
          else
            er1 = e1/ep1
          endif

*      -- output
          write(*,'(1x,i2,g24.16,2g16.8,2f8.4)') 
     &          nc, r1, e0, e1, er0, er1

*      -- check convergence
*      -- require at least 'itmin' iterations and error < 'ermin'.
          if(j.gt.itmin .and. abs(e1).lt.ermin) then
            if(er1.eq.0) then
              ans = r1
              ncc = nc

*          -- skip subsequent iterations
              goto 900

*        -- now error is growing ?
            else if(er1.ge.1) then
              ans = rp1
              ncc = nc-1

*          -- skip subsequent iterations
              goto 900
            endif
          endif

*      -- the result when iterations are terminated.
          ans = r1

*    -- end of iterations
        enddo

  900   continue
*    -- result
        write(*,'(1x,i2,g24.16)') ncc, ans

*    -- comparison with the analytic result.
        write(*,'(i3, g24.16, a, a, g12.4)') 
     &        npl, tv, ' (analytic)', ' err=', abs((tv-ans)/tv) 
      enddo

      end
*=======================================================================
      subroutine integt(nc, r, fnc)
*
*     Input
*       nc  : the previous value of the iteration number 
*       r   : previous result when nc >= 0.
*       fnc : function to be integrated
*     Output
*       r   : result of the iteration
*       nc  : updated iteration number

      implicit none
      integer  nc
      double precision r
      double precision fnc

      double precision zero, one, two, half
      parameter (zero = 0, one = 1, two = 2, half = one/two)

      double precision dn, s, x
      integer          np, j
*-----------------------------------------------------------------------

*  -- the first call
      if(nc.lt.0) then
        nc = 0
        r  = half*(fnc(zero) + fnc(one))
c       write(*,*) nc, x, r
        return
      endif

*  -- the number of points
      np  = 2**nc

*  -- width
      dn  = half/dble(np)

*  -- for the next call
      nc  = nc + 1

*  -- summation of the values of function 'fnc(x)'
      s = 0
      do j = 1, np
        x = (two*j-one)*dn
        s = s + fnc(x)
      enddo

*  -- the result
      r = half*r + dn*s

c     write(*,*) nc, x, r 
      return
      end
*=======================================================================
      double precision function fncl(x)
*
*     fncl(x) = x**npx * log(x)**npl       for npl >= 0, x >= 0
*
*     Input
*       x : real number for which the function is calculated
*       npx, npl : selector of a function.  This parameter is passed
*                  through common block /idat/
*     Output
*       calculated function value

      implicit none

*  -- argument of the function
      double precision x

*  -- parameters
      integer    npx, npl
      common    /idat/npx, npl
*-----------------------------------------------------------------------

*  -- check 
      if(x.lt.0) then
        write(*,*) 'log(x), x = ',x,' < 0'
        stop 999
      endif

*    -- special point x=0 ( x*log(x)=0 for x=0 )
      if(npx.gt.0 .and. x.eq.0) then
        fncl = 0
      else
        fncl = x**npx * log(x)**npl
      endif

      return
      end
*=======================================================================
      double precision function fnce(x)
*
*     fnce(x) = x**npx * exp(npl * x)      
*
*     Input
*       x : real number for which the function is calculated
*       npx, npl : selector of a function.  This parameter is passed
*                  through common block /idat/
*     Output
*       calculated function value

      implicit none

*  -- argument of the function
      double precision x

*  -- parameters
      integer    npx, npl
      common    /idat/npx, npl
*-----------------------------------------------------------------------
*  -- func = Power * exp
      fnce = x**npx * exp(npl * x)

      return
      end
*=======================================================================
      double precision function fncla()
*
*                        /1
*     analytic result of | dx fncl(x)
*                        /0
*
      implicit none

*  -- parameters
      integer    npx, npl
      common    /idat/npx, npl

*  -- local vars
      double precision d1, a1, t1
      integer          k
*-----------------------------------------------------------------------
      d1 = - 1/dble(npl)
      a1 = d1
      t1 = a1
      do k = 1, npx
        a1 = a1 * (npx-k+1) * d1
        t1 = t1 + a1
      enddo
      fncla = - t1*exp(dble(npl)) + a1

      return
      end
*=======================================================================
      double precision function fncea()
*
*                        /1
*     analytic result of | dx fnce(x)
*                        /0
*
      implicit none

*  -- parameters
      integer    npx, npl
      common    /idat/npx, npl

*  -- local vars
      double precision fpx, tv
      integer          k
*-----------------------------------------------------------------------
      fpx = 1/dble(npx+1)
      tv = fpx
      do k = 1, npl
        tv = tv * (- k)*fpx
      enddo
      fncea = tv

      return
      end
